@@ -44,7 +44,8 @@ cplus_expand_expr (exp, target, tmode, modifier)
   rtx original_target = target;
   int ignore = target == const0_rtx;
 
-  if (ignore) target = 0, original_target = 0;
+  if (ignore)
+    target = 0, original_target = 0;
 
   /* No sense saving up arithmetic to be done
      if it's all in the wrong mode to form part of an address.
@@ -74,7 +75,9 @@ cplus_expand_expr (exp, target, tmode, modifier)
 	tree type = TREE_TYPE (exp), slot;
 	tree fn_type = TREE_TYPE (TREE_TYPE (func));
 	tree return_type = TREE_TYPE (fn_type);
+	tree call_exp;
 	rtx call_target, return_target;
+	int pcc_struct_return = 0;
 
 	/* The expression `init' wants to initialize what
 	   `target' represents.  SLOT holds the slot for TARGET.  */
@@ -118,17 +121,53 @@ cplus_expand_expr (exp, target, tmode, modifier)
 	  }
 	else
 	  {
+#ifdef PCC_STATIC_STRUCT_RETURN
+	    pcc_struct_return = 1;
+	    call_target = 0;
+#else
 	    call_target = target;
+#endif
 	  }
 	if (call_target)
-	  preserve_temp_slots (call_target);
+	  {
+	    preserve_temp_slots (call_target);
+
+	    /* Make this a valid memory address now.  The code below assumes
+	       that it can compare rtx and make assumptions based on the
+	       result.  The assumptions are true only if the address was
+	       valid to begin with.  */
+	    call_target = validize_mem (call_target);
+	  }
+
 	preserve_temp_slots (DECL_RTL (slot));
-	return_target = expand_expr (build (CALL_EXPR, type, func, args, 0), call_target, mode, 0);
+	call_exp = build (CALL_EXPR, type, func, args, 0);
+	TREE_SIDE_EFFECTS (call_exp) = 1;
+	return_target = expand_expr (call_exp, call_target, mode, 0);
 	free_temp_slots ();
 	if (call_target == 0)
-	  call_target = return_target;
-	else if (call_target != return_target)
 	  {
+	    if (pcc_struct_return)
+	      {
+		tree init = build (RTL_EXPR, type, 0, return_target);
+		TREE_ADDRESSABLE (init) = 1;
+		expand_aggr_init (slot, init, 0);
+		if (TYPE_NEEDS_DESTRUCTOR (type))
+		  {
+		    init = build (RTL_EXPR, build_reference_type (type), 0,
+				  XEXP (return_target, 0));
+		    init = maybe_build_cleanup (convert_from_reference (init));
+		    if (init != NULL_TREE)
+		      expand_expr (init, 0, 0, 0);
+		  }
+		call_target = return_target = DECL_RTL (slot);
+	      }
+	    else
+	      call_target = return_target;
+	  }
+
+	if (call_target != return_target)
+	  {
+	    my_friendly_assert (! TYPE_NEEDS_CONSTRUCTING (type), 317);
 	    if (GET_MODE (return_target) == BLKmode)
 	      emit_block_move (call_target, return_target, expr_size (exp),
 			       TYPE_ALIGN (type) / BITS_PER_UNIT);
@@ -159,10 +198,17 @@ cplus_expand_expr (exp, target, tmode, modifier)
 
     case OFFSET_REF:
       {
+#if 1
+	return expand_expr (default_conversion (resolve_offset_ref (exp)),
+			    target, tmode, EXPAND_NORMAL);
+#else
+	/* This is old crusty code, and does not handle all that the
+	   resolve_offset_ref function does.  (mrs) */
 	tree base = build_unary_op (ADDR_EXPR, TREE_OPERAND (exp, 0), 0);
 	tree offset = build_unary_op (ADDR_EXPR, TREE_OPERAND (exp, 1), 0);
 	return expand_expr (build (PLUS_EXPR, TREE_TYPE (exp), base, offset),
 			    target, tmode, EXPAND_NORMAL);
+#endif
       }
 
     default:
